#include <stdio.h>
#include <ctype.h>
#include <string>

#include "common.h"
#include "Patcher.h"



Patcher::Patcher()
   :
   mBase("Base bitmap"),
   mPatch("Patch bitmap")
   {
   }
   

int Patcher::runScript(const string& sFileName)
   {
   int iRetVal = 0;
   
   try
      {
      printf("Reading patch script: %s...\n", sFileName.c_str());
      mScr.open(sFileName);

      while (mScr.getLine())
         {
         string sVerb = mScr.readWord();

         int i;
         for (i = 0; i < mnVocab; ++i)
            {
            const Vocab& vocab = mVocab[i];
            if (sVerb == vocab.szVerb)
               {
               string sObj = (*vocab.szObj ? string(vocab.szObj) : mScr.readWord());
               if (sObj.empty())
                  throw Exception("Missing arguments.");
               (this->*(vocab.pmfnCmd))(sObj);
               break;
               }
            }

         if (i == mnVocab)
            throw Exception("Syntax error.");
         }

      puts("Done.");
      }
   catch (const Exception& e)
      {
      if (mScr.opened())
         {fprintf(stderr, "Line %ld: %s", mScr.lineNo(), mScr.line().c_str());}
      fprintf(stderr, "**** %s\n", e.c_str());
      iRetVal = 1;
      }

   mPatch.tm.unload();
   mBase.tm.unload();
   mScr.close();

   return iRetVal;
   }


Patcher::Bitmap& Patcher::bitmap(const string& sWhich)
   {
   if (sWhich == "base")
      return mBase;
   else if (sWhich == "patch")
      return mPatch;
   else
      throw Exception("Invalid argument: " + sWhich);
   }
   

void Patcher::load(const string& sWhich)
   {
   TileMap& tm = bitmap(sWhich).tm;
   const string& sFileName = mScr.rest();
   if (sFileName.empty())
      throw Exception("Filename not specified.");
   
   printf("Loading %s bitmap: %s...\n", sWhich.c_str(), sFileName.c_str());
   tm.load(sFileName);
   }


void Patcher::save(const string&)
   {
   const string& sFileName = mScr.rest();
   if (sFileName.empty())
      throw Exception("Filename not specified.");
   
   FILE* fp = fopen(sFileName.c_str(), "rb");
   if (fp)
      {
      fclose(fp);
      fprintf(stderr, "%s already exists: overwrite (y/n)? ",
              sFileName.c_str());
      char c = getchar(); while (getchar() != '\n');
      if (tolower(c) != 'y') return;
      }
   
   printf("Saving output bitmap: %s...\n", sFileName.c_str());
   mBase.tm.saveAs(sFileName);
   }


void Patcher::tileSize(const string&)
   {
   long lWidth = 1, lHeight = 1;
   int n = sscanf(mScr.rest().c_str(), "%ld%ld", &lWidth, &lHeight);
   if (n < 1)
      throw Exception("Missing arguments.");
   else if (n == 1)
      {lHeight = lWidth;}
   
   mBase.ts.w = mPatch.ts.w = lWidth;
   mBase.ts.h = mPatch.ts.h = lHeight;
   mBase.setSpecs();
   mPatch.setSpecs();
   }


void Patcher::tileSpecs(const string& sWhich)
   {
   Bitmap& bm = bitmap(sWhich);
   TileSpecs& ts = bm.ts;
   
   sscanf(mScr.rest().c_str(), "%ld%ld%ld%ld%ld%ld",
          &ts.dx0, &ts.dy0, &ts.dx1, &ts.dy1, &ts.dx2, &ts.dy2);
   bm.setSpecs();
   }


void Patcher::merge(const string& sWhat)
   {
   bool tiles = false;
   if (sWhat == "tiles")
      {tiles = true;}
   else if (sWhat != "pixels")
      throw Exception("Invalid argument: " + sWhat);

   string sOp = mScr.readWord();
   if (sOp.length() != 1)
      throw Exception("Invalid operator: " + sOp);
   static const string sOps = "=&|^";
   size_t p = sOps.find(sOp[0]);
   if (p == string::npos)
      throw Exception("Invalid operator: " + sOp);
   TileMap::Operation op = (TileMap::Operation)p;
  
   long dx=0, dy=0, sx=0, sy=0, w=1, h=1; int m=1, n=1;
   int nArgs = sscanf(mScr.rest().c_str(), "%ld%ld%ld%ld%ld%ld%d%d",
                      &dx, &dy, &sx, &sy, &w, &h, &m, &n);
   switch (nArgs)    //no breaks
     {
     case 0: dx = 0;
     case 1: dy = 0;
     case 2: sx = dx;
     case 3: sy = dy;
     case 4: w = (tiles ? 1 : mBase.ts.w);
     case 5: h = (tiles ? 1 : mBase.ts.h);
     case 6: m = 1;
     case 7: n = 1;
     }

   int i, j; long x, y;
   for (j=0, y=dy; j < n; ++j, y+=h)
      {
      for (i=0, x=dx; i < m; ++i, x+=w)
         {
         if (tiles)
            {mBase.tm.mergeTileRange(op, x, y, mPatch.tm, sx, sy, w, h);}
         else
            {mBase.tm.mergePixelRange(op, x, y, mPatch.tm, sx, sy, w, h);}
         }
      }
   }


const Patcher::Vocab Patcher::mVocab[] =
{
   {"load",     "",            &Patcher::load},
   {"save",     "base",        &Patcher::save},
   {"tilesize", "both",        &Patcher::tileSize},
   {"tilespecs","",            &Patcher::tileSpecs},
   {"merge",    "",            &Patcher::merge},
   {"<p",       "pixels",      &Patcher::merge},
   {"<t",       "tiles",       &Patcher::merge}
};

const int Patcher::mnVocab = sizeof(Patcher::mVocab) / sizeof(Patcher::Vocab);

