#include <stdlib.h>
#include <stdio.h>

#include "common.h"
#include "TileMap.h"



TileMap::TileMap(const string& sName)
   :
   msName(sName),
   mLoaded(false)
   {
   }


void TileMap::raiseError(const string& s) const
   {
   throw Exception(msName + ": " + s);
   }

void TileMap::assertLoaded() const
   {
   if (!mLoaded)
      raiseError("Not loaded.");
   }


void TileMap::load(const string& sFileName)
   {
   unload();
   FILE* fp = NULL;

try
{
   fp = fopen(sFileName.c_str(), "rb");
   if (!fp)
      raiseError("Error opening file.");
      
   fseek(fp, 0, SEEK_END);
   mlFileSize = ftell(fp);
   rewind(fp);
   
   if (mlFileSize < 0x36)
      raiseError("Invalid bitmap file.");
   
   if (!mpFileData.alloc(mlFileSize))
      raiseError("Not enough memory.");
      
   if (fread(mpFileData, mlFileSize, 1, fp) != 1)
      raiseError("Error reading file.");

   fclose(fp); fp = NULL;
      
   if (!(mpFileData[0]=='B' && mpFileData[1]=='M'))
      raiseError("Unrecognized bitmap format.");
   
   dword dwCompression = *(dword*)(mpFileData+0x1E);
   if (dwCompression != 0)
      raiseError("Compressed bitmaps not supported.");
   
   int iBPP = *(word*)(mpFileData+0x1C);
   if (!(iBPP == 24  ||  iBPP == 8))
      raiseError("Only 24-BPP and 8-BPP bitmaps supported.");
   miBytesPP = iBPP >> 3;
      
   mlWidth = *(dword*)(mpFileData+0x12);
   mlHeight = *(dword*)(mpFileData+0x16);
   mlLineSize = ((mlWidth*miBytesPP + 4-1) / 4) * 4;

   mlBmpOffset = *(dword*)(mpFileData+0x0A);
   if ((mlBmpOffset + mlHeight*mlLineSize) > mlFileSize)
      raiseError("Invalid bitmap file.");

   mLoaded = true;
}
catch (...)
{
   if (fp) fclose(fp);
   mpFileData.release();
   throw;
}
   }


void TileMap::unload()
   {
   if (mLoaded)
      {
      mpFileData.release();
      mLoaded = false;
      }
   }


void TileMap::setTileSpecs(const TileSpecs& ts)
   {
   mts = ts;
   }


void TileMap::saveAs(const string& sFileName) const
   {
   assertLoaded();

   FILE* fp = fopen(sFileName.c_str(), "wb");
   if (!fp)
      raiseError("Error creating file.");
   size_t n = fwrite(mpFileData, mlFileSize, 1, fp);
   fclose(fp);
   
   if (n != 1)
      raiseError("Error writing file.");
   }


void TileMap::mergePixelRange(Operation op, long lDestX, long lDestY,
                              const TileMap& tmSrc, long lSrcX, long lSrcY,
                              long lWidth, long lHeight)
   {
   assertCompatibleWith(tmSrc, false);
      
   checkPixelRange(lDestX, lDestY, lWidth, lHeight);
   tmSrc.checkPixelRange(lSrcX, lSrcY, lWidth, lHeight);

   MergeFunc* const pMergeFunc = mpMergeFunc[op];
   byte* pDest = mpFileData + calcPixelOffset(lDestX, lDestY);
   const byte* pSrc = tmSrc.mpFileData + tmSrc.calcPixelOffset(lSrcX, lSrcY);
   for (long l = 0; l < lHeight; ++l)
      {
      (*pMergeFunc)(pDest, pSrc, lWidth*miBytesPP);
      pDest -= mlLineSize;
      pSrc -= tmSrc.mlLineSize;
      }
 }
   

void TileMap::mergeTileRange(Operation op, int iDestCol, int iDestRow,
                             const TileMap& tmSrc, int iSrcCol, int iSrcRow,
                             int iCols, int iRows)
   {
   assertCompatibleWith(tmSrc, true);
      
   int nc = iCols, nr = iRows;
   long w = mts.w, h = mts.h;
   if (mts.dx1==0 && mts.dx2==0 && tmSrc.mts.dx1==0 && tmSrc.mts.dx2==0)
      {w *= nc; nc = 1;}
   if (mts.dy1==0 && mts.dy2==0 && tmSrc.mts.dy1==0 && tmSrc.mts.dy2==0)
      {h *= nr; nr = 1;}   
   
   for (int r = 0; r < nr; ++r)
      {
      for (int c = 0; c < nc; ++c)
         {
         long lDestX, lDestY;
         calcTileCoords(iDestCol+c, iDestRow+r, &lDestX, &lDestY);
         long lSrcX, lSrcY;
         tmSrc.calcTileCoords(iSrcCol+c, iSrcRow+r, &lSrcX, &lSrcY);
         mergePixelRange(op, lDestX, lDestY, tmSrc, lSrcX, lSrcY, w, h);
         }
      }
   }
   

void TileMap::checkPixelCoords(long x, long y) const
   {
   if (!(x >= 0  &&  x < mlWidth  &&  y >= 0  &&  y < mlHeight))
      raiseError("Co-ordinates out of range.");
   }
   
void TileMap::checkPixelRange(long x, long y, long w, long h) const
   {
   checkPixelCoords(x, y);
   checkPixelCoords(x+w-1, y+h-1);
   }
  

void TileMap::calcTileCoords(int iCol, int iRow, long* px, long* py) const
   {
   *px = mts.dx0 + iCol*mts.blockWidth() + mts.dx1;
   *py = mts.dy0 + iRow*mts.blockHeight() + mts.dy1;
   }
   

long TileMap::calcPixelOffset(long x, long y) const
   {
   return mlBmpOffset  +  ((mlHeight-1-y) * mlLineSize) + (x * miBytesPP);
   }


void TileMap::assertCompatibleWith(const TileMap& tm, bool tiles) const
   { 
   assertLoaded();
   tm.assertLoaded();
   
   if (!(miBytesPP == tm.miBytesPP) ||
       (tiles  &&  !(mts.w==tm.mts.w && mts.h==tm.mts.h)))
      raiseError("Bitmaps are incompatible.");
   }
   

#define DEFINE_MERGEFUNC(f, op) \
void* TileMap::mem##f(void* pDest, const void* pSrc, size_t n) \
   { \
   byte* pd = (byte*)pDest; \
   const byte* ps = (const byte*)pSrc; \
   for (size_t i = 0; i < n; ++i, ++pd, ++ps) \
      { \
      *pd op##= *ps; \
      } \
   return pDest; \
   }

DEFINE_MERGEFUNC(AND, &)
DEFINE_MERGEFUNC(OR , |)
DEFINE_MERGEFUNC(XOR, ^)

TileMap::MergeFunc* const TileMap::mpMergeFunc[] =
   {
   &memcpy,
   &TileMap::memAND,
   &TileMap::memOR,
   &TileMap::memXOR
   };
